#ifndef MEMORY_ALLOCATOR_H
#define MEMORY_ALLOCATOR_H

#include <stdio.h>

/**
 * Homemade memory allocator that dynamically allocates a pool of memory upon
 * startup.  It then issues memory from that pool and reclaims space when it
 * is freed.  Perfect allocation can be performed since only one type of object
 * is stored in each type of allocator.
 *
 * WARNING: Items allocated or deleted will not have constructor/destructor
 *          called.  Should really just be used for classes/types without 
 *          internal dynamic allocation.
 *
 * @author Scott Schaefer
 */

template <class T>
class MemoryAllocator
{
	protected:
		/// the array to use for allocations
		T* array;
		/// next array to keep track of the next item in the list
		/// (linked list emulation)
		int* next;
		/// the first element in the free list
		int top;
		/// the total number of items held in this allocator
		int length;

	public:
		/**
		 * Constructor.
		 */
		MemoryAllocator ( int num )
		{
			top = 0;
			array = new T [ num ];
			next = new int [ num ];
			length = num;

			for ( int i = 0; i < num; i++ )
			{
				next [ i ] = i + 1;
			}
			next [ num - 1 ] = -1;
		}

		/** 
		 * Destructor.
		 */
		~MemoryAllocator ( void )
		{
			if ( array != NULL )
			{
				delete[] array;
			}
			if ( next != NULL )
			{
				delete[] next;
			}
		}

		/**
		 * Gets the unique address associated with this item.  This address is
		 * zero based.  It's essentially &item - &base.
		 *
		 * @param item the item to find the address of
		 *
		 * @return the address of the specified item
		 */
		int getAddress ( const T*& item )
		{
			int address;

			// stupid check
			if ( item == NULL )
			{
				return -1;
			}

			// calculate address in array (we know that we have a linear block of memory )
			// don't need to divide by sizeof ( T ) because 
			// pointer arithmetic already does it for you
			address = ( item - array );

			// another stupid check
			if ( address >= length || address < 0 )
			{
				return -1;
			}

			return address;
		}

		/**
		 * Allocates an item from this allocator.
		 *
		 * @return the address of the allocated item or NULL if no more data
		 *         is available for allocation
		 */
		T* allocateItem ( void )
		{
			int prevTop;

			if ( top == -1 )
			{
				printf ( "Out of memory!\n" );
				// out of memory!
				return NULL;
			}

			prevTop = top;

			top = next [ top ];
			next [ prevTop ] = -1;

			return &(array [ prevTop ]);
		}

		/**
		 * Frees an item back into the memory pool.  We're cool in that we only
		 * need the address of the item being freed and can use getAddress to
		 * determine its position in the array.
		 *
		 * @param item the item to free
		 */
		void freeItem ( const T*& item )
		{
			int address = getAddress ( item );

			// stupid check
			if ( address == -1 )
			{
				return;
			}

			if ( next [ address ] != -1 )
			{
				// error, trying to free something that wasn't allocated
				printf ( "ERROR: Tried to free address %d, but it wasn't allocated\n", address );
				return;
			}

			if ( top == -1 )
			{
				top = address;
				return;
			}

			next [ address ] = top;
			top = address;
		}
};
#endif